import * as Electron from 'electron';

import { Type, ObjectUtil } from '@ucap/core';
import { AppChannel } from '@ucap/electron-core';

import { Builder as AppBuilder } from '../api/app/builder';
import { AppConfiguration } from '../api/app/models/app-configuration';
import { getAppSettings } from '../api/app/decorators/app-settings';
import { Builder as BrowserWindowBuilder } from '../api/browser-window/builder';
import { ElectronBrowserWindow } from '../api/browser-window/electron-browser-window';

import { ServiceBuilder } from './service-builder';
import { LifecycleHooks, getHook } from './hooks';

import { Attacher } from './attacher';

export const CLASS_NAME = 'ucapClassName';

export class Platform {
  protected static instanceMap: Map<string, any> = new Map();
  protected static instance: Platform;

  protected startedAt = new Date();
  protected appConfig: AppConfiguration;
  protected _app: any;

  static getInstance<T>(typeOfInstance: Type<any>): T {
    return Platform.instanceMap.get(Platform.getClassName(typeOfInstance)) as T;
  }

  static getAppConfig(): AppConfiguration {
    return Platform.instance.appConfig;
  }

  static buildBrowserWindow<T extends ElectronBrowserWindow>(
    browserWindowType: Type<T>
  ): T {
    const win = BrowserWindowBuilder.build(
      Platform.instance.appConfig,
      browserWindowType
    );
    Platform.setInstance(browserWindowType, win);
    Platform.inject(win);
    Attacher.attach(win);

    return win;
  }

  static destroyBrowserWindow<T extends ElectronBrowserWindow>(
    browserWindow: T
  ) {
    BrowserWindowBuilder.destroy(browserWindow);
    Platform.removeInstance(browserWindow);
    Attacher.dettach(browserWindow);
  }

  static async bootstrap(
    app: Type<any>,
    configuration: Partial<AppConfiguration> = {}
  ): Promise<Platform> {
    const _configuration = getAppSettings(app, configuration);

    Platform.instance = new Platform();
    Platform.instance.appConfig = _configuration;

    Platform.instance._app = AppBuilder.build(app, _configuration);
    Platform.setInstance(app, Platform.instance._app);

    if (!!_configuration.services && 0 < _configuration.services.length) {
      _configuration.services.forEach((serviceType) => {
        const i = ServiceBuilder.build(serviceType, _configuration);
        Platform.setInstance(serviceType, i);
      });
    }

    Platform.instanceMap.forEach((i, n) => {
      Platform.inject(i);
    });

    let afterInitHook: (...args: any[]) => any;
    let afterInitHookReturn: Promise<any> | undefined;

    Platform.instanceMap.forEach(async (i, n) => {
      afterInitHook = getHook(i, LifecycleHooks.AfterInit);
      if (!!afterInitHook) {
        afterInitHookReturn = afterInitHook.call(i);
        if (afterInitHookReturn instanceof Promise) {
          await Promise.resolve(afterInitHookReturn);
        }
      }
    });

    Platform.instanceMap.forEach((i, n) => {
      Attacher.attach(i);
    });

    await Platform.instance.runLifecycle();

    return Platform.instance;
  }

  private static getClassName(typeOfInstance: Type<any>): string {
    const name = typeOfInstance[CLASS_NAME];
    if (!name) {
      throw Error(
        `name of class is not defined. use static property ${CLASS_NAME}`
      );
    }

    return name;
  }

  private static setInstance(typeOfInstance: Type<any>, instance: any) {
    const name = typeOfInstance[CLASS_NAME];
    if (Platform.instanceMap.has(name)) {
      throw Error(`duplicated name of class[${name}]`);
    }

    Platform.instanceMap.set(name, instance);
  }

  private static removeInstance(instance: any) {
    const name = ObjectUtil.getConstructor(instance)[CLASS_NAME];
    if (!Platform.instanceMap.has(name)) {
      return;
    }

    Platform.instanceMap.delete(name);
  }

  private static inject(instance: any) {
    Platform.instanceMap.forEach((i, n) => {
      instance[n] = i;
    });
  }

  async start(): Promise<void> {}

  private async runLifecycle(): Promise<void> {
    // tslint:disable-next-line: variable-name
    const __this = this;
    Electron.app.once(AppChannel.beforeQuit, async () => {
      let beforeDestroyHook: (...args: any[]) => any;
      let beforeDestroyHookReturn: Promise<any> | undefined;

      Platform.instanceMap.forEach(async (i, n) => {
        beforeDestroyHook = getHook(i, LifecycleHooks.AfterInit);
        if (!!beforeDestroyHook) {
          beforeDestroyHookReturn = beforeDestroyHook.call(i);
          if (beforeDestroyHookReturn instanceof Promise) {
            await Promise.resolve(beforeDestroyHookReturn);
          }
        }
      });
    });
  }
}
